/*************************************************************************

         name: acceptabilityConstraints.pl (Chapter 11)
      version: June 18, 1998
  description: Consistency and Informativity
      authors: Patrick Blackburn & Johan Bos
 
*************************************************************************/

:- module(acceptabilityConstraints,[consistentReadings/2,
				    informativeReadings/2,
				    localConsistentReadings/2,
				    localInformativeReadings/2]).

:- use_module(drs2fol,[drs2fol/2]),
   use_module(comsemOperators),
   use_module(comsemPredicates,[append/3]),
   use_module(mergeDRT,[mergeDrs/2]),
   use_module(callTheoremProver,[callTheoremProver/3]),
   use_module(semOntology,[generateOntology/1]).

/*========================================================================
   Select Readings 
========================================================================*/

informativeReadings([],[]).
informativeReadings([(DrsSoFar,Drs)|Potential],[Drs|Selected]):-
   informative(DrsSoFar,Drs), !,
   informativeReadings(Potential,Selected).
informativeReadings([_|Potential],Selected):-
   informativeReadings(Potential,Selected).

consistentReadings([],[]).
consistentReadings([(Old,Drs)|Potential],[(Old,Drs)|Selected]):-
   consistent(Drs), !,
   consistentReadings(Potential,Selected).
consistentReadings([_|Potential],Selected):-
   consistentReadings(Potential,Selected).

localConsistentReadings([],[]).
localConsistentReadings([Drs|Potential],[Drs|Selected]):-
   localConsistent(Drs), !,
   localConsistentReadings(Potential,Selected).
localConsistentReadings([_|Potential],Selected):-
   localConsistentReadings(Potential,Selected).

localInformativeReadings([],[]).
localInformativeReadings([Drs|Potential],[Drs|Selected]):-
   localInformative(Drs), !,
   localInformativeReadings(Potential,Selected).
localInformativeReadings([_|Potential],Selected):-
   localInformativeReadings(Potential,Selected).

/*========================================================================
   Informativity
========================================================================*/

informative(drs([],[]),_):-!.
informative(OldDrs,NewDrs):-
   backgroundKnowledge(Chi),
   drs2fol(OldDrs,Phi),
   drs2fol(NewDrs,Psi),
   callTheoremProver(Chi,~(Phi > Psi),Proof),
   (Proof=yes, !, fail; true).

/*========================================================================
   Consistency
========================================================================*/

consistent(NewDrs):-
   backgroundKnowledge(Chi),
   drs2fol(NewDrs,Phi),
   callTheoremProver(Chi,Phi,Proof),
   (Proof=yes, !, fail; true).

/*========================================================================
   Local Informativity
========================================================================*/

localInformative(Drs):-
   findall((Super,Sub),superSubDrs(Drs,drs([],[])-Super,Sub),List),
   allLocalInformative(List).

allLocalInformative([]).
allLocalInformative([(Super,Sub)|Others]):-
   backgroundKnowledge(Chi),
   drs2fol(drs([],[Super>Sub]),Phi),
   callTheoremProver(Chi,~Phi,Proof),
   (Proof=yes, !, fail; allLocalInformative(Others)).

/*========================================================================
   Local Consistency
========================================================================*/

localConsistent(Drs):-
   findall((Super,Sub),superSubDrs(Drs,drs([],[])-Super,Sub),List),
   allLocalConsistent(List).

allLocalConsistent([]).
allLocalConsistent([(Super,Sub)|Others]):-
   backgroundKnowledge(Chi),
   drs2fol(drs([],[Super>drs([],[~Sub])]),Phi),
   callTheoremProver(Chi,~Phi,Proof),
   (Proof=yes, !, fail; allLocalConsistent(Others)).

/*========================================================================
   Background Knowledge
========================================================================*/

backgroundKnowledge(Formulas):-
   knowledge(Formulas1),
   generateOntology(Formulas2),
   append(Formulas1,Formulas2,Formulas).

knowledge([
           forall(X,forall(Y,have(X,Y)>of(Y,X))),
           forall(X,forall(Y,of(Y,X)>have(X,Y))),
           forall(X,female(X)&married(X)>exists(Y,husband(Y)&have(X,Y))),
           forall(X,forall(Y,husband(Y)&have(X,Y)>married(X)&female(X)))
	  ]).

/*========================================================================
   Compute super- and subordinated DRSs
========================================================================*/

superSubDrs(drs(D,[Sub > _|C]),Drs-Super,Sub):-
   mergeDrs(merge(Drs,drs(D,C)),Super).
superSubDrs(drs(D,[B > Sub|C]),Drs-Super,Sub):-
   mergeDrs(merge(merge(drs(D,C),B),Drs),Super).
superSubDrs(drs(D,[B1 > B2|C]),Drs-Super,Sub):-
   superSubDrs(B2,merge(Drs,merge(merge(drs(D,C),B1),B2))-Super,Sub).
superSubDrs(drs(D,[Sub v _|C]),Drs-Super,Sub):-
   mergeDrs(merge(Drs,drs(D,C)),Super).
superSubDrs(drs(D,[_ v Sub|C]),Drs-Super,Sub):-
   mergeDrs(merge(Drs,drs(D,C)),Super).
superSubDrs(drs(D,[B v _|C]),Drs-Super,Sub):-
   superSubDrs(B,merge(Drs,merge(drs(D,C),B))-Super,Sub).
superSubDrs(drs(D,[_ v B|C]),Drs-Super,Sub):-
   superSubDrs(B,merge(Drs,merge(drs(D,C),B))-Super,Sub).
superSubDrs(drs(D,[~ Sub|C]),Drs-Super,Sub):-
   mergeDrs(merge(Drs,drs(D,C)),Super).
superSubDrs(drs(D,[~ B|C]),Drs-Super,Sub):-
   superSubDrs(B,merge(Drs,merge(drs(D,C),B))-Super,Sub).
superSubDrs(drs(D,[Cond|C]),Drs-Super,Sub):-
   superSubDrs(drs([],C),Drs-B,Sub),
   mergeDrs(merge(drs(D,[Cond]),B),Super).
