/*************************************************************************

         name: freeVarTabl.pl (Chapter 5)
      version: Dec 10, 1998
  description: Free Variable Semantic Tableaux
               Uses a number of ideas from Melvin Fitting's 
               implementation of an unsigned tableaux theorem prover 
               for first-order logic, in "First-Order Logic and 
               Automated Theorem Proving", Second Edition (1996),
               Graduate Texts in Computer Science, Springer.
               For more details, see the Notes to Chapter 5.
      authors: Patrick Blackburn & Johan Bos
 
*************************************************************************/

:- module(freeVarTabl,[valid/2,saturate/2,notatedFormula/3]).

:- use_module(comsemOperators),
   use_module(comsemPredicates,[member/2,removeFirst/3,unify/2,append/3,
                                compose/3,newFunctionCounter/1,
                                substitute/4]).

/*========================================================================

   General Tableaux Predicates
   ---------------------------

saturate(+Tableau,+Qdepth) 
   Expand Tableau until it is closed, allowing Qdepth 
   applications of the universal rule.

closed(+Tableau) 
   Every branch of Tableau can be made to contain a contradiction, 
   after a suitable free variable substitution. 

skolemFunction(+VarList,-SkoTerm) 
   VarList is a list of free variables, and SkoTerm is a previously 
   unused Skolem function symbol fun(N) applied to those free variables.

valid(+X,+Qdepth) 
   Try to create a tableau expansion for f(X) that is closed allowing 
   Qdepth applications of the universal rule. 

notatedFormula(Notated,Free,SignedFormula) 
   Notated is a notated formula, Free its associated free variable 
   list, and SignedFormula the signed formula part.

========================================================================*/

saturate(Tableau,_Q):-
   closed(Tableau).

saturate(OldTableau,Qdepth):-
   expand(OldTableau,Qdepth,NewTableau,NewQdepth),!,
   saturate(NewTableau,NewQdepth).

closed([]).
closed([Branch|Rest]):-
   member(NotatedOne,Branch),
   notatedFormula(NotatedOne,_,t(X)),
   member(NotatedTwo,Branch),
   notatedFormula(NotatedTwo,_,f(Y)),
   unify(X,Y),
   closed(Rest).

skolemFunction(VarList,SkolemTerm) :-
   newFunctionCounter(N),
   compose(SkolemTerm,fun,[N|VarList]).

valid(X,Qdepth):-
   notatedFormula(NotatedFormula,[],f(X)),
   saturate([[NotatedFormula]],Qdepth).

notatedFormula(n(Free,Formula),Free,Formula).

/*========================================================================

   Tableaux Expansion Predicates
   ---------------------------

expand(+Oldtableau,+OldQdepth,-Newtableau,-NewQdepth) 
   Newtableaux with Q-depth NewQdepth is the result of applying  
   a tableaux expansion rule to Oldtableaux with a Q-depth of OldQdepth.

unaryExpansion(+Branch,-NewBranch)
   Take Branch as input, and return NewBranches if a tableau rule
   allows unary expansion.

conjunctiveExpansion(+Branch,-NewBranch)
   Take Branch as input, and return the NewBranch if a tableau rule 
   allows conjunctive expansion. 

disjunctiveExpansion(+Branch,-NewBranch1,-NewBranch2)
   Take Branch as input, and return the NewNranch1 and NewBranch2 
   if a tableau rule allows disjunctive expansion. 

existentialExpansion(+Branch,-NewBranch)
   Take Branch as input, and return the NewBranch if a tableau rule 
   allows existential expansion. 

universalExpansion(+Branch,+OldQDepth,-NewBranch,-NewQDepth)
   Take Branch and OldQD as input, and return the NewBranch and 
   NewQDepthif a tableau rule allow universal expansion.

========================================================================*/

expand([Branch|Tableau],QD,[NewBranch|Tableau],QD):-
   unaryExpansion(Branch,NewBranch).

expand([Branch|Tableau],QD,[NewBranch|Tableau],QD):-
   conjunctiveExpansion(Branch,NewBranch).

expand([Branch|Tableau],QD,[NewBranch|Tableau],QD):-
   existentialExpansion(Branch,NewBranch).

expand([Branch|Tableau],QD,[NewBranch1,NewBranch2|Tableau],QD):-
   disjunctiveExpansion(Branch,NewBranch1,NewBranch2).

expand([Branch|Tableau],OldQD,NewTableau,NewQD):-
   universalExpansion(Branch,OldQD,NewBranch,NewQD),
   append(Tableau,[NewBranch],NewTableau).

expand([Branch|Rest],OldQD,[Branch|Newrest],NewQD):-
   expand(Rest,OldQD,Newrest,NewQD).

unaryExpansion(Branch,[NotatedComponent|Temp]) :-
   unary(SignedFormula,Component),
   notatedFormula(NotatedFormula,Free,SignedFormula),
   removeFirst(NotatedFormula,Branch,Temp),
   notatedFormula(NotatedComponent,Free,Component).

conjunctiveExpansion(Branch,[NotatedComp1,NotatedComp2|Temp]):-
   conjunctive(SignedFormula,Comp1,Comp2),
   notatedFormula(NotatedFormula,Free,SignedFormula),
   removeFirst(NotatedFormula,Branch,Temp),
   notatedFormula(NotatedComp1,Free,Comp1),
   notatedFormula(NotatedComp2,Free,Comp2).

disjunctiveExpansion(Branch,[NotComp1|Temp],[NotComp2|Temp]):-
   disjunctive(SignedFormula,Comp1,Comp2),
   notatedFormula(NotatedFormula,Free,SignedFormula),
   removeFirst(NotatedFormula,Branch,Temp),
   notatedFormula(NotComp1,Free,Comp1),
   notatedFormula(NotComp2,Free,Comp2).

existentialExpansion(Branch,[NotatedInstance|Temp]):-
   notatedFormula(NotatedFormula,Free,SignedFormula),
   existential(SignedFormula),
   removeFirst(NotatedFormula,Branch,Temp),
   skolemFunction(Free,Term),
   instance(SignedFormula,Term,Instance),
   notatedFormula(NotatedInstance,Free,Instance).

universalExpansion(Branch,OldQD,NewBranch,NewQD):-
   OldQD > 0, NewQD is OldQD - 1,
   member(NotatedFormula,Branch),
   notatedFormula(NotatedFormula,Free,SignedFormula),
   universal(SignedFormula),
   removeFirst(NotatedFormula,Branch,Temp),
   instance(SignedFormula,V,Instance),
   notatedFormula(NotatedInstance,[V|Free],Instance),
   append([NotatedInstance|Temp],[NotatedFormula],NewBranch).

/*========================================================================

   Formula Identification
   ----------------------

conjunctive(?F,?Comp1,?Comp2)
   F is a conjunctive signed formula with components Comp1 and Comp2

disjunctive(?F,?Comp1,?Comp2) 
   F is a disjunctive signed formula with components Comp1 and Comp2

unary(?F,?Comp) 
   F is a signed formula with component Comp

universal(?F) 
   F is a universal formula.

existential(?F) 
   F is an existential formula.

instance(F,Term,Ins)
   F is a signed quantified formula, and Ins is the result of 
   removing the quantifier and replacing all free occurrences of 
   the quantified variable by occurrences of Term.

========================================================================*/

conjunctive(t(X & Y),t(X),t(Y)).
conjunctive(f(X v Y),f(X),f(Y)).
conjunctive(f(X > Y),t(X),f(Y)).

disjunctive(f(X & Y),f(X),f(Y)).
disjunctive(t(X v Y),t(X),t(Y)).
disjunctive(t(X > Y),f(X),t(Y)).

unary(t(~X),f(X)).
unary(f(~X),t(X)). 

universal(t(forall(_,_))).
universal(f(exists(_,_))).

existential(t(exists(_,_))).
existential(f(forall(_,_))).

instance(t(forall(X,F)),Term,t(NewF)):- 
   substitute(Term,X,F,NewF).
instance(f(exists(X,F)),Term,f(NewF)):- 
   substitute(Term,X,F,NewF).
instance(t(exists(X,F)),Term,t(NewF)):- 
   substitute(Term,X,F,NewF).
instance(f(forall(X,F)),Term,f(NewF)):- 
   substitute(Term,X,F,NewF).

