/**********************************************************************

         name: mainFocusDRT.pl (Chapter 9)
      version: Feb 3, 1999
  description: Pronoun Resolution with the Focusing Algorithm
      authors: Patrick Blackburn & Johan Bos
 
**********************************************************************/

:- module(mainFocusDRT,[parse/0,resolveDrs/2]).

:- use_module(readLine,[readLine/1]),
   use_module(comsemOperators),
   use_module(comsemPredicates,[printRepresentation/1,simpleTerms/1,
                                compose/3,append/3,member/2]),
   use_module(betaConversion,[betaConvert/2]),
   use_module(bindingDRT,[potentialAntecedent/3,properBinding/3]).

:- [englishGrammar], [englishLexicon], [semMacrosLambdaDRT].

/*========================================================================
   Driver Predicate
========================================================================*/

parse:-
   parse(drs([],[],[],[])).

parse(Old):-
   readLine(Discourse),
   d(SDrs,Discourse,[]),
   betaConvert(SDrs,Drs),
   resolveDrs([merge(Old,Drs)]-A,[]-Anaphora),
   updateFocus(Anaphora,A,[New]),
   printRepresentation(New), !,
   parse(New).

/*=====================================================================
    Pronoun Resolution: DRSs
=====================================================================*/

resolveDrs([merge(B1,B2)|A1]-[drs(D,AF,DF,C)|A5],R1-R3):-
   resolveDrs([B1|A1]-A2,R1-R2),	
   updateFocus(R2,A2,A3),
   resolveDrs([B2|A3]-A4,R2-R3),
   updateFocus(R3,A4,[drs(D1,AF1,DF1,C1),drs(D2,AF2,DF2,C2)|A5]),
   append(D1,D2,D),
   append(AF1,AF2,AF),
   append(DF1,DF2,DF),
   append(C1,C2,C).

resolveDrs([drs(D,AF,DF,C)|A1]-A3,R1-R2):-
   resolveConds(C,[drs(D,AF,DF,[])|A1]-A2,R1-R2),
   updateFocus(R2,A2,A3).

resolveDrs([drs(D,C)|A1]-A3,R1-R2):-
   resolveConds(C,[drs(D,[],[],[])|A1]-A2,R1-R2),
   updateFocus(R2,A2,A3).

/*=====================================================================
    Pronoun Resolution: if pronoun appears as agent, then choose the
    actor focus as preferred antecedent
=====================================================================*/

resolveDrs([alfa(X,Type,Gender,B1)|A1]-A2,R1-[Type:X|R2]):-
   role(agent,X,B1), !,
   (
    currentActorFocus(A1,AF),
    potentialAntecedent(A1,X,Gender),
    X == AF,
    properBinding(Type,X,B1)
   ;   
    currentDiscourseFocus(A1,DF),
    potentialAntecedent(A1,X,Gender),
    X == DF,
    properBinding(Type,X,B1)
   ;
    potentialAntecedent(A1,X,Gender),
    properBinding(Type,X,B1)
   ), 
   resolveDrs([B1|A1]-A2,R1-R2).


/*=====================================================================
    Pronoun Resolution: if pronoun appears as non-agent, then choose 
    the discourse focus as preferred antecedent
=====================================================================*/

resolveDrs([alfa(X,Type,Gender,B1)|A1]-A2,R1-[Type:X|R2]):-
   (
    currentDiscourseFocus(A1,DF),
    potentialAntecedent(A1,X,Gender),
    X == DF,
    properBinding(Type,X,B1)
   ;   
    currentActorFocus(A1,AF),
    potentialAntecedent(A1,X,Gender),
    X == AF,
    properBinding(Type,X,B1)
   ;
    potentialAntecedent(A1,X,Gender),
    properBinding(Type,X,B1)
   ),
   resolveDrs([B1|A1]-A2,R1-R2). 


/*=====================================================================
    Pronoun Resolution: DRS-conditions
=====================================================================*/

resolveConds([~B1|Conds],A1-A3,R1-R3):-
   resolveDrs([B1|A1]-[B2,drs(D,AF,DF,C)|A2],R1-R2),
   resolveConds(Conds,[drs(D,AF,DF,[~B2|C])|A2]-A3,R2-R3).

resolveConds([B1 > B2|Conds],A1-A5,R1-R4):-
   resolveDrs([B1|A1]-A2,R1-R2),
   updateFocus(R2,A2,A3), 
   resolveDrs([B2|A3]-[B4,B3,drs(D,AF,DF,C)|A4],R2-R3),
   resolveConds(Conds,[drs(D,AF,DF,[B3 > B4|C])|A4]-A5,R3-R4).

resolveConds([B1 v B2|Conds],A1-A4,R1-R4):-
   resolveDrs([B1|A1]-[B3|A2],R1-R2),
   resolveDrs([B2|A2]-[B4,drs(D,AF,DF,C)|A3],R2-R3),
   resolveConds(Conds,[drs(D,AF,DF,[B3 v B4|C])|A3]-A4,R3-R4).

resolveConds([Basic|Conds],[drs(D,AF,DF,C)|A1]-A2,R1-R2):-
   compose(Basic,_Symbol,Arguments),
   simpleTerms(Arguments),
   resolveConds(Conds,[drs(D,AF,DF,[Basic|C])|A1]-A2,R1-R2).

resolveConds([],A-A,R-R).


/*=====================================================================
     Focus Update
=====================================================================*/

updateFocus(Anaphora,DRSs1,Updated):-
   updateActorFocus(DRSs1,DRSs2),
   updateDiscourseFocus(Anaphora,DRSs2,Updated).

/*=====================================================================
     Update Actor Focus
=====================================================================*/

updateActorFocus([drs(D,AF,DF,C)|A],Updated):-
   expectedActorFocus(drs(D,C),Focus), !,
   addActorFocus(Focus,[drs(D,AF,DF,C)|A],Updated).

updateActorFocus(DRSs,DRSs).

/*=====================================================================
     Update Discourse Focus
=====================================================================*/

updateDiscourseFocus(Anaphora,DRSs,Updated):-
   currentDiscourseFocus(DRSs,Current), 
   member(_:A,Anaphora), A==Current, !,
   Updated=DRSs.

updateDiscourseFocus(_Anaphora,[drs(D,AF,DF,C)|A],Updated):-
   expectedDiscourseFocus(drs(D,C),Focus), !,
   addDiscourseFocus(Focus,[drs(D,AF,DF,C)|A],Updated).

updateDiscourseFocus(_,DRSs,DRSs).


/*=====================================================================
     Add new focus to DRS
=====================================================================*/


addActorFocus(Focus,DRSs,Updated):-
   (
       currentActorFocus(DRSs,Current),
       Current==Focus, !,
       Updated=DRSs
   ;
       DRSs=[drs(D,AF,DF,C)|A],
       Updated=[drs(D,[Focus|AF],DF,C)|A]
   ).

addDiscourseFocus(Focus,DRSs,Updated):-
   (
       currentDiscourseFocus(DRSs,Current),
       Current==Focus, !,
       Updated=DRSs
   ;
       DRSs=[drs(D,AF,DF,C)|A],
       Updated=[drs(D,AF,[Focus|DF],C)|A]
   ).


/*=====================================================================
     Current Foci
=====================================================================*/

currentDiscourseFocus(DRSs,Focus):-
   member(drs(_,_,[Focus|_],_),DRSs), !.

currentActorFocus(DRSs,Focus):-
   member(drs(_,[Focus|_],_,_),DRSs), !.

/*=====================================================================
     Expected Focus
=====================================================================*/

expectedActorFocus(drs(Dom,Conds),Focus):-
   member(Focus,Dom),
   role(agent,Focus,Conds).

expectedDiscourseFocus(drs(Dom,Conds),Focus):-
   member(Focus,Dom),
   role(theme,Focus,Conds).


/*=====================================================================
     Thematic Roles
=====================================================================*/

role(Role,X,drs(_,_,_,Conds)):-
   role(Role,X,Conds).

role(Role,X,drs(_,Conds)):-
   role(Role,X,Conds).

role(Role,X,merge(B1,B2)):-
   role(Role,X,B1);
   role(Role,X,B2).

role(Role,X,alfa(_,_,_,B)):-
   role(Role,X,B).

role(Role,X,[~ B|_]):-
   role(Role,X,B).

role(Role,X,[B1 > B2|_]):-
   role(Role,X,B1);
   role(Role,X,B2).

role(Role,X,[B1 v B2|_]):-
   role(Role,X,B1);
   role(Role,X,B2).

role(Role,X,[_|Conds]):-
   role(Role,X,Conds).

role(agent,X,[C|_]):-
   agent(X,C).

role(theme,X,[C|_]):-
   theme(X,C).

/*=====================================================================
     Agent Position
=====================================================================*/

agent(X,C):-
   compose(C,Sym,[Y,_]),
   X==Y,
   lexicon(tv,Sym,_,inf).


/*=====================================================================
     Theme Position
=====================================================================*/

theme(X,C):-
   compose(C,Sym,[Y]),
   X==Y,
   lexicon(iv,Sym,_,inf).

theme(X,C):-
   compose(C,Sym,[_,Y]),
   X==Y,
   lexicon(tv,Sym,_,inf).

theme(X,C):-
   compose(C,'=',[Y,Z]),
   var(Z),
   X==Y.

