/*************************************************************************

         name: propTabl.pl (Chapter 4)
      version: Dec 3, 1998
  description: Propositional Tableaux Program
      authors: Patrick Blackburn & Johan Bos
 
*************************************************************************/

:- module(propTabl,[valid/1,saturate/1]).

:- use_module(comsemOperators),
   use_module(comsemPredicates,[member/2,removeFirst/3]).

/*========================================================================

   General Tableau Predicates
   --------------------------

saturate(+Tableau) 
   Expand Tableau until it is closed.

closed(+Tableau) 
   Every branch of Tableau contains a contradiction.

valid(+Formula) 
   Try to create a closed tableau expansion for f(Formula). 

========================================================================*/

saturate(Tableau):- 
   closed(Tableau).

saturate(OldTableau):-
   expand(OldTableau,NewTableau),
   saturate(NewTableau).

closed([]).

closed([Branch|Rest]):-
    member(t(X),Branch),
    member(f(X),Branch),
    closed(Rest).

valid(F):-
   saturate([[f(F)]]).

/*========================================================================

   Tableau Expansion Predicates
   ----------------------------

expand(+Oldtableau,-Newtableau) 
   Newtableaux is the result of applying a tableaux expansion 
   rule to Oldtableaux.

unaryExpansion(+Branch,-NewBranch)
   Take Branch as input, and return NewBranches if a tableau rule
   allows unary expansion.

conjunctiveExpansion(+Branch,-NewBranch)
   Take Branch as input, and return the NewBranch if a tableau rule 
   allows conjunctive expansion. 

disjunctiveExpansion(+Branch,-NewBranch1,-NewBranch2)
   Take Branch as input, and return the NewNranch1 and NewBranch2 
   if a tableau rule allows disjunctive expansion. 

========================================================================*/

expand([Branch|Tableau],[NewBranch|Tableau]):-
   unaryExpansion(Branch,NewBranch).

expand([Branch|Tableau],[NewBranch|Tableau]):-
   conjunctiveExpansion(Branch,NewBranch).

expand([Branch|Tableau],[NewBranch1,NewBranch2|Tableau]):-
   disjunctiveExpansion(Branch,NewBranch1,NewBranch2).

expand([Branch|Rest],[Branch|Newrest]):-
   expand(Rest,Newrest).

unaryExpansion(Branch,[Component|Temp]) :-
   unary(SignedFormula,Component),
   removeFirst(SignedFormula,Branch,Temp).

conjunctiveExpansion(Branch,[Comp1,Comp2|Temp]):-
   conjunctive(SignedFormula,Comp1,Comp2),
   removeFirst(SignedFormula,Branch,Temp).

disjunctiveExpansion(Branch,[Comp1|Temp],[Comp2|Temp]):-
   disjunctive(SignedFormula,Comp1,Comp2),
   removeFirst(SignedFormula,Branch,Temp).

/*========================================================================

   Formula Identification
   ----------------------

conjunctive(?F,?Comp1,?Comp2)
   F is a conjunctive signed formula with components Comp1 and Comp2.

disjunctive(?F,?Comp1,?Comp2) 
   F is a disjunctive signed formula with components Comp1 and Comp2.

unary(?F,?Comp) 
   F is a signed formula with component Comp.

========================================================================*/

conjunctive(t(X & Y),t(X),t(Y)).
conjunctive(f(X v Y),f(X),f(Y)).
conjunctive(f(X > Y),t(X),f(Y)).

disjunctive(f(X & Y),f(X),f(Y)).
disjunctive(t(X v Y),t(X),t(Y)).
disjunctive(t(X > Y),f(X),t(Y)).

unary(t(~X),f(X)).
unary(f(~X),t(X)). 
